from .convolution import log_gabor
from .decomposition import (eigenvalue_decomposition,
                            principal_component_decomposition)
from .linalg import dot_inplace_left, dot_inplace_right
