
# This file was generated by 'versioneer.py' (0.10) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.4.0'
version_full = '04277b9c50dafb8d51773d84b40b9582c166e31d'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

